/*
 * Created on 02.10.2004
 */
package weka.classifiers.ucd.ensemble;

import java.io.Serializable;
import java.util.*;

import weka.classifiers.*;
import weka.classifiers.bayes.NaiveBayes;
import weka.classifiers.functions.SMO;
import weka.core.*;
import weka.filters.unsupervised.attribute.*;

/**
 * @author andreas
 */
public class FeatureSpaceDivider extends Classifier implements OptionHandler, Serializable {
	
	public double classifyInstance(Instance i) throws Exception {

		// classify with all classifiers
		detailedResults = classifyWithBiasedClassifiers(i);

		detailedResults[clc.length] = arbiter[0].classifyInstance(i);
		detailedResults[clc.length+1] = -1;
		detailedResults[clc.length+2] = -1;

		int nClasses = clc.length;
		
		// if the agreement ratio is less than minAgreement,
		// ask the arbiter
		
		Vote vote = vote(detailedResults);
		
		if (talkative) System.out.println("Agreement: " + vote.agreement);
		if (vote.agreement < minAgreement) {
			// ask arbiter
			if (talkative) System.out.println("Asking arbiter");
			detailedResults[clc.length+1] = detailedResults[clc.length];
			return detailedResults[clc.length];
		} else {
			detailedResults[clc.length+2] = (double) vote.votedClass;
			if ((double) vote.votedClass != detailedResults[0] && talkative && nClasses == 2) {
				System.out.println("Whoops! vote.votedClass=" + vote.votedClass + ", cls=" + detailedResults[0]);
			}
			return (double) vote.votedClass;
		}
	}

	/**
	 * some additional statistics
	 * element 0-(nClasses-1): individual classifiers
	 * element nClasses: arbiter
	 * element nClasses+1: arbiter (when actually used)
	 * element nClasses+2: -1 if disagree, = element 0 otherwise 
	 */
	double detailedResults[];
	
	/**
	 * Instances for subclassifiers
	 */
	Instances[] lc;
	
	/**
	 * Subclassifiers
	 */
	Classifier[] clc;
	
	/**
	 * Subclassifier
	 */
	Classifier[] arbiter;
	
	/**
	 * The mode of operation
	 */
	int method = 2;
	
	/**
	 * If rounds > 1, then the arbiter will be replaced with another
	 * FeatureSpaceDivider with the same parameters as this one except
	 * that rounds will be decreased by 1.
	 */
	int rounds = 1; 

	/**
	 * Auxiliary class to store the result of a voting of
	 * (the biased) classifiers
	 */	
	private class Vote {
		public double agreement;
		public int votedClass;
		public Vote(double _agreement, int _votedClass) {
			agreement = _agreement;
			votedClass = _votedClass;
		}
	}
	
	public static final int METHOD_UNDERSAMPLING = 1;
	public static final int METHOD_OVERSAMPLING = 2;
	public static final int METHOD_TRAININGERROR = 3;
	public static final Tag[] TAGS_METHOD = {
			new Tag(METHOD_UNDERSAMPLING, "Random undersampling"),
			new Tag(METHOD_OVERSAMPLING, "Weighted oversampling"),
			new Tag(METHOD_TRAININGERROR, "Training-error based (similar to stacking)")
	};

	/**
	 * Whether or not to produce additional output
	 */
	boolean talkative = false;
	
	/**
	 * If this is true, all instances that are inserted into
	 * the arbiter classifier get a weight of 4.
	 * 
	 * As stupid as it sounds, but this can boost accuracy by
	 * about 5% !!
	 * 
	 * However, by default, this option is off, so that we cannot
	 * be accused of cheating.
	 */
	boolean smoCheat = false;
	
	/**
	 * Whether or not to use log based weighting instead of fixed weights/probabilities
	 */
	boolean logBasedWeighting = false;
	
	/**
	 * If true, use a 1-vs-all approach when handling multi-class problems 
	 */
	boolean oneVsAll = false;
	
	/**
	 * The fraction of classifiers that have to agree, or else the
	 * arbiter classifier will be asked.
	 */
	double minAgreement = 1.0;
	
	/**
	 * per-class probabilities for undersampling
	 */
	double undersamplingProbability = 0.1;
	
	/**
	 * weight-multiplicator for oversampling
	 */
	double oversamplingWeight = 3; // good in general
	
	/**
	 * Base classifier class name (default Naive Bayes)
	 */
	// String classifier = "weka.classifiers.bayes.NaiveBayes";
	Classifier classifier = new NaiveBayes();
	
	/**
	 * Options for base classifier
	 */
	// String[] classifierOptions;
	
	/**
	 * Random number seed (for random undersampling)
	 */
	int seed = 1;
	
	/**
	 * @param detailedResults The output of classifyWithBiasedClassifiers
	 * @return A Vote consisting of agreement ratio and the majority class
	 */
	private Vote vote(double[] detailedResults) {
		
		int nClasses = clc.length;

		if (oneVsAll) {
			
			// In the one-vs-all case, there is only either total agreement
			// (agreement=1) or disagreement (agreement=0). In the latter case,
			// votedClass is set to -1 (as it is useless anyway when
			// agreement=0). 
			
			// The classifiers have been trained on a dataset that was created
			// by a MakeIndicator filter, i.e. class index 0 means negative and
			// class index 1 means positive.
			
			int classification = -1;

			// loop all one-vs-all classifiers
			for (int n = 0; n < nClasses; n++) {
				int predictedClass = (int) detailedResults[n];
				if (predictedClass == 1) {	// classified as positive
					if (classification == -1) {
						classification = n;
					} else {
						// disagreement
						return new Vote(0, -1);
					}
				}
			}
			if (classification == -1) {
				// disagreement, ask arbiter
				return new Vote(0, -1);
			}
			// agreement
			detailedResults[clc.length+1] = (double) classification;
			return new Vote(1, classification);

		} else {
		
			// vote and find the class with max votes
				
			int[] nVotes = new int[nClasses]; 
			for (int n = 0; n < clc.length; n++) {
				nVotes[n] = 0;
			}
			int argmax = -1;
			int max = 0;
			for (int n = 0; n < nClasses; n++) {
				int predictedClass = (int) detailedResults[n];
				if (predictedClass != -1) {
					nVotes[predictedClass]++;
					if (nVotes[predictedClass] > max) {
						max = nVotes[predictedClass];
						argmax = predictedClass;
					}
				}
			}
			double agreement = (double) max / (double) nClasses;
			return new Vote(agreement, argmax);
		}
	}
	
	/**
	 * Classify a single instance with all biased classifiers
	 * @param i A single instance that will be classifierd
	 * @return An array of detailed results with room left for the arbiter
	 * @throws Exception
	 */
	private double[] classifyWithBiasedClassifiers(Instance i) throws Exception {

		// Reserve array for results of individual classifiers,
		// leaving room for arbiter.
		
		double[] detailedResults = new double[clc.length+3];

		// classify with all biased classifiers
		
		for (int n = 0; n < clc.length; n++) {
			try {
				detailedResults[n] = clc[n].classifyInstance(i);
			} catch (Exception e) {
				detailedResults[n] = -1;
				if (talkative) {
					System.err.println("Whoops! (Something strange happened in classifier " + n + ".)");
					e.printStackTrace();
				}
			}
		}
		return detailedResults;
	}
	
	/* (non-Javadoc)
	 * @see weka.classifiers.Classifier#buildClassifier(weka.core.Instances)
	 */
	public void buildClassifier(Instances trainingSet) throws Exception {
		Random rnd = new Random(seed); // seed
		int nClasses = trainingSet.numClasses();
		int nInstances = trainingSet.numInstances();
		lc = new Instances[nClasses];
		clc = new Classifier[nClasses];

		int[] nInstancesPerClass = new int[nClasses];

		if (logBasedWeighting) {
			for (int m = 0; m < nInstances; m++) {
				Instance i = trainingSet.instance(m);
				nInstancesPerClass[(int) i.classValue()]++;
			}
		}
		
		switch (method) {
			
			///////////////////////////////////////////////////////////////////
			// 	METHOD_TRAININGERROR
			//
					
			
			case METHOD_TRAININGERROR:

				// Create a simple classifier first
				// This "pre-classifier" is just for finding out the
				// training set mistakes. With this information, biased
				// classifiers for each class are created with all instances *except*
				// the training set mistakes that belong to other classes
				
				Classifier preClassifier[] = Classifier.makeCopies(classifier, 1);
				preClassifier[0].buildClassifier(trainingSet);
				
				// initialize datasets for biased classifier
		
				for (int n = 0; n < nClasses; n++) {
					lc[n] = new Instances(trainingSet, 0);
				}
		
				// Compute training set errors
				
				int trainingErrors = 0;
				
				// This strategy seems to work best, at least with SVM and
				// the diabetes dataset:
				// For each biased classifier:
				// Instances of that class with training error are omitted
				
				for (int m = 0; m < nInstances; m++) {
					Instance i = trainingSet.instance(m);
					int trueClass = (int) i.classValue();
					int preClassifierSays = (int) preClassifier[0].classifyInstance(i);
					
					// add to classifier for true class only if correct
					
					if (trueClass == preClassifierSays) {
						// lc[trueClass].add(new Instance(i));
						// lc[trueClass].add((Instance) i.copy());
						lc[trueClass].add(i);
					} else {
						trainingErrors++;
					}
					
					// add to all classifiers except for the true class
					
					for (int n = 0; n < nClasses; n++) {
						if (n != trueClass) {
							// lc[n].add(new Instance(i));
							// lc[n].add((Instance) i.copy());
							lc[n].add(i);
						}
					}
				}
						
				if (talkative) System.out.println("Training errors: " + trainingErrors);
				
				break;

			///////////////////////////////////////////////////////////////////
			// 	METHOD_OVERSAMPLING
			//

			case METHOD_OVERSAMPLING:
			
				// Create a biased classifier for each class
				// by overweighting positive instances
				// (they are inserted twice)
				
	
				for (int n = 0; n < nClasses; n++) {
					
					if (logBasedWeighting) {
						int sum = 0;
						for (int a = 0; a < nClasses; a++) {
							if (a != n) {
								sum += nInstancesPerClass[a];
							}
						}
						oversamplingWeight = Math.log(sum);
					}
					
					lc[n] = new Instances(trainingSet, 0);
					for (int m = 0; m < nInstances; m++) {
						Instance i = trainingSet.instance(m); 
						if ((trainingSet.instance(m).classValue() == (double) n)) {
							i = (Instance) i.copy();
							i.setWeight(oversamplingWeight);
						}
						lc[n].add(i);
					}
				}
				break;
				
				
			///////////////////////////////////////////////////////////////////
			// 	METHOD_UNDERSAMPLING
			//

			case METHOD_UNDERSAMPLING:
	
				// Create a biased classifier for each class
				// by omitting negative instances
				// (negative instances are inserted with probability p)
				
				for (int n = 0; n < nClasses; n++) {
	
					if (logBasedWeighting) {
						undersamplingProbability = 1.0 / Math.log(nInstancesPerClass[n]);
					}
	
					lc[n] = new Instances(trainingSet, 0);
					for (int m = 0; m < nInstances; m++) {
						Instance i = trainingSet.instance(m); 
						if ((i.classValue() != (double) n)) {
							if (rnd.nextDouble() < undersamplingProbability) {
								// insert a copy of the instance
								// lc[n].add(new Instance(i));
								// lc[n].add((Instance) i.copy());
								lc[n].add(i);
							}
						} else {
							// lc[n].add(new Instance(i));
							// lc[n].add((Instance) i.copy());
							lc[n].add(i);
						}
					}
				}
				break;
				
			default:
				throw new Exception("Mode not correctly specified!");
		}

		// Create a biased classifier for each class
			
		clc = Classifier.makeCopies(classifier, nClasses);
		for (int n = 0; n < nClasses; n++) {
			
			// if one-vs-all, re-label instances for the biased classifiers
			
			if (oneVsAll) {
				MakeIndicator classFilter = new MakeIndicator();
				classFilter.setValueIndex(n);
				classFilter.setNumeric(false);
				classFilter.setInputFormat(lc[n]);
				Instances newInsts = weka.filters.Filter.useFilter(lc[n], classFilter);
				lc[n] = newInsts;				
			}
			
			clc[n].buildClassifier(lc[n]);
		}

		// Arbiter is trained on instances where biased classifiers disagree
		// (or agree, but on the wrong class)
		
		Instances t = new Instances(trainingSet, 0);
		for (int m = 0; m < nInstances; m++) {
			Instance i = trainingSet.instance(m);
			int trueClass = (int) i.classValue();
			double[] details = classifyWithBiasedClassifiers(i);
			Vote v = vote(details);
			
			if (v.agreement < minAgreement || v.votedClass != trueClass) {

				if (smoCheat) {

					// Inserting one and the same instance multiple times helps SMO
					// (What the heck...?)

					// Instance j = new Instance(i);
					Instance j = (Instance) i.copy();
					j.setWeight(4);
					t.add(j);
				} else {
					t.add(i);
				}
				
				
				
			}
			
//			double c = i.classValue();
//			for (int n = 0; n < nClasses; n++) {
//				if (clc[n].classifyInstance(i) != c) {
//					t.add(i);
//					break;
//				}
//			}
		}
		
		if (rounds > 1) {
			arbiter = Classifier.makeCopies(this, 1);
			((FeatureSpaceDivider) arbiter[0]).setRounds(rounds - 1);
		} else {
			arbiter = Classifier.makeCopies(classifier, 1);
		}
		if (t.numInstances() == 0) {
			// If the biased classifiers agreed on every instance (!),
			// then train the arbiter on the complete training set.
			// (What else could we do!?)
			if (talkative) {
				System.out.println("100% agreement, building arbiter on full training set.");
			} 
			arbiter[0].buildClassifier(trainingSet);
		} else {
			// build arbiter normally
			if (talkative) {
				System.out.println("Building arbiter on " + t.numInstances() + " instances.");
			}
			arbiter[0].buildClassifier(t);
		}
	}

	/* (non-Javadoc)
	 * @see weka.core.OptionHandler#listOptions()
	 */
	public Enumeration listOptions() {
		Vector options = new Vector(5);
		options.add(
				new Option("\tSpecify the full class name of a base classifier " +
						"(default NaiveBayes).", "W", 1, "-W <classname>"));
		options.add(
				new Option("\tSets the method to use. Valid values are " +
						"1 (random undersampling), 2 (oversampling), " +
						"3 (training error) (default 1)", "M", 1, "-M <int>"));
		options.add(
				new Option("\tRandom number seed when using undersampling",
						"S", 1, "-Q <int>"));
		options.add(
				new Option("\tInstance weight when using oversampling",
						"H", 1, "-H <double>"));
		options.add(
				new Option("\tToggle debug output (0 off or 1 on, default 0)",
						"D", 1, "-D <int>"));
		options.add(
				new Option("\tProbability for random undersampling",
						"P", 1, "-P <double>"));
		options.add(
				new Option("\tAutomatically set weights based on the log of the " +
						"number of instances in each class (default 0)",
						"L", 1, "-L <int>"));
		options.add(
				new Option("\tUse 1-vs-All mode for multiclass classification (default 0)",
						"C", 1, "-C <int>"));
		options.add(
				new Option("\tMinimum agreement of classifiers for not asking arbiter (default 1.0)",
						"A", 1, "-A <double>"));
		options.add(
				new Option("\tUse \"SMO Cheat\": Training instances for arbiter are overweighted (default 0)",
						"X", 1, "-X <int>"));
		options.add(
				new Option("\tSpecify the number of rounds for this algorithm. If > 1, the arbiter " +
						"classifier will be replaced with another FeatureSpaceDivider.",
						"R", 1, "-R <int>"));
		return options.elements();
	}

	/* (non-Javadoc)
	 * @see weka.core.OptionHandler#setOptions(java.lang.String[])
	 */
	public void setOptions(String[] options) throws Exception {
		
		String classifierNameOption = Utils.getOption('W', options);
		
		String methodOption = Utils.getOption('M', options);
		if (methodOption.length() != 0) {
			method = Integer.parseInt(methodOption);
		}
		
		String seedOption = Utils.getOption('S', options);
		if (seedOption.length() != 0) {
			seed = Integer.parseInt(seedOption);
		}
		
		String weightOption = Utils.getOption('H', options);
		if (weightOption.length() != 0 ) {
			oversamplingWeight = Double.parseDouble(weightOption);
		}
		
		String debugOption = Utils.getOption('D', options);
		if (debugOption.length() != 0) {
			int debugOptionInt = Integer.parseInt(debugOption);
			if (debugOptionInt == 1) talkative = true;
		}
		
		String probOption = Utils.getOption('P', options);
		if (probOption.length() != 0) {
			undersamplingProbability = Double.parseDouble(probOption);
		}
		
		String multiClassOption = Utils.getOption('C', options);
		if (multiClassOption.length() != 0) {
			int multiClassOptionInt = Integer.parseInt(multiClassOption);
			if (multiClassOptionInt == 1) oneVsAll = true;
		}
		
		String logBasedOption = Utils.getOption('L', options);
		if (logBasedOption.length() != 0) {
			int logBasedOptionInt = Integer.parseInt(logBasedOption);
			if (logBasedOptionInt == 1) logBasedWeighting = true;
		}

		String smoCheatOption = Utils.getOption('X', options);
		if (smoCheatOption.length() != 0) {
			int smoCheatOptionInt = Integer.parseInt(smoCheatOption);
			if (smoCheatOptionInt == 1) smoCheat = true;
		}

		String minAgreeOption = Utils.getOption('A', options);
		if (minAgreeOption.length() != 0) {
			minAgreement = Double.parseDouble(minAgreeOption);
		}

		String roundsOption = Utils.getOption('R', options);
		if (roundsOption.length() != 0) {
			rounds = Integer.parseInt(roundsOption);
		}

		String[] classifierOptions = Utils.partitionOptions(options);
		if (classifierNameOption.length() != 0) {
			classifier = Classifier.forName(classifierNameOption, classifierOptions);
		}
	}

	/* (non-Javadoc)
	 * @see weka.core.OptionHandler#getOptions()
	 */
	public String[] getOptions() {
	    String [] classifierOptions = new String [0];
	    if ((classifier != null) &&
		(classifier instanceof OptionHandler)) {
	      classifierOptions = ((OptionHandler) classifier).getOptions();
	    }
	    // String [] options = new String [classifierOptions.length + 13];
		String [] options = new String [classifierOptions.length + 23];
		int c = 0;
		
		options[c++] = "-W"; options[c++] = classifier.getClass().getName();
		options[c++] = "-M"; options[c++] = "" + method;
		options[c++] = "-P"; options[c++] = "" + undersamplingProbability;
		options[c++] = "-S"; options[c++] = "" + seed;
		options[c++] = "-H"; options[c++] = "" + oversamplingWeight;
		options[c++] = "-A"; options[c++] = "" + minAgreement;
		options[c++] = "-R"; options[c++] = "" + rounds;
		if (talkative) {
			options[c++] = "-D"; options[c++] = "1";
		} else {
			options[c++] = "-D"; options[c++] = "0";
		}
		if (smoCheat) {
			options[c++] = "-X"; options[c++] = "1";
		} else {
			options[c++] = "-X"; options[c++] = "0";
		}
		if (logBasedWeighting) {
			options[c++] = "-L"; options[c++] = "1";
		} else {
			options[c++] = "-L"; options[c++] = "0";
		}
		if (oneVsAll) {
			options[c++] = "-C"; options[c++] = "1";
		} else {
			options[c++] = "-C"; options[c++] = "0";
		}
		options[c++] = "--";
		
	    System.arraycopy(classifierOptions, 0, options, c, 
			     classifierOptions.length);
	    c += classifierOptions.length;
	    while (c < options.length) {
	      options[c++] = "";
	    }

		return options;
	}

	public Classifier getClassifier() {
		return classifier;
	}
	public void setClassifier(Classifier classifier) {
		this.classifier = classifier;
	}

	public boolean isTalkative() {
		return talkative;
	}
	public void setTalkative(boolean talk) {
		this.talkative = talk;
	}
	public SelectedTag getMethod() {
		return new SelectedTag(method, TAGS_METHOD);
	}
	public void setMethod(SelectedTag newMethod) {
		if (newMethod.getTags() == TAGS_METHOD) {
			method = newMethod.getSelectedTag().getID();
		}
	}
	public String seedTipText() {
		return "Used as seed for random number generator if random undersampling is selected.";
	}
	public int getSeed() {
		return seed;
	}
	public void setSeed(int seed) {
		this.seed = seed;
	}
	public double[] getDetailedResults() {
		return detailedResults;
	}
	public String oversamplingWeightTipText() {
		return "Weight that is assigned to positive instances in oversampling mode";
	}
	public double getOversamplingWeight() {
		return oversamplingWeight;
	}
	public void setOversamplingWeight(double oversamplingWeight) {
		this.oversamplingWeight = oversamplingWeight;
	}
	public String undersamplingProbabilityTipText() {
		return "Probability for adding negative instances in undersampling mode";
	}
	public double getUndersamplingProbability() {
		return undersamplingProbability;
	}
	public void setUndersamplingProbability(double undersamplingProbability) {
		this.undersamplingProbability = undersamplingProbability;
	}
	public String globalInfo() {
		return "An ensemble method that trains specialised classifiers on " +
				"subsets of instances so that effectively the feature space is " +
				"split into sections. (Andreas Hess, University College Dublin," +
				" <andreas.hess@ucd.ie>)";
	}
	/**
	 * @return
	 */
	public boolean isLogBasedWeighting() {
		return logBasedWeighting;
	}

	/**
	 * @return
	 */
	public double getMinAgreement() {
		return minAgreement;
	}

	/**
	 * @return
	 */
	public boolean isOneVsAll() {
		return oneVsAll;
	}

	/**
	 * @param b
	 */
	public void setLogBasedWeighting(boolean b) {
		logBasedWeighting = b;
	}

	/**
	 * @param d
	 */
	public void setMinAgreement(double d) {
		minAgreement = d;
	}

	/**
	 * @param b
	 */
	public void setOneVsAll(boolean b) {
		oneVsAll = b;
	}

	/**
	 * @return
	 */
	public boolean isSmoCheat() {
		return smoCheat;
	}

	/**
	 * @param b
	 */
	public void setSmoCheat(boolean b) {
		smoCheat = b;
	}

	/**
	 * @return
	 */
	public int getRounds() {
		return rounds;
	}

	/**
	 * @param i
	 */
	public void setRounds(int i) {
		rounds = i;
	}

	/**
	 * Main method for testing this class.
	 * Note that if you call the classifier this way, the default settings
	 * are SMO and random undersampling (as opposed to NaiveBayes and
	 * overweighting).
	 */
	public static void main(String[] argv) {

	  FeatureSpaceDivider scheme;

	  try {
		scheme = new FeatureSpaceDivider();
		scheme.setClassifier(new SMO());
		scheme.setMethod(
			new SelectedTag(FeatureSpaceDivider.METHOD_UNDERSAMPLING,
			FeatureSpaceDivider.TAGS_METHOD));
		scheme.setRounds(2);
		scheme.setUndersamplingProbability(0.2);
		System.out.println(Evaluation.evaluateModel(scheme, argv));
	  } catch (Exception e) {
		e.printStackTrace();
		System.err.println(e.getMessage());
	  }
	}


}
